// c't 3x4x5 puzzle solver
// (C) 2003 by Johannes Overmann <Johannes.Overmann@gmx.de>
//
// $Id: ctpuzzle.cc,v 1.18 2003/04/30 02:11:39 root Exp $

#include <vector>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <unistd.h>
using namespace std;


// Sorry for the lack of documentation and for the unclean code, but its
// 2003-04-30 03:45 in the morning ;-)


// Warning: Different coordinate convention than most other participants:
// ======================================================================
// X is 5 units long
// Y is 4 units long
// Z is 3 units long


// some low level types
typedef unsigned long long U64;
typedef unsigned int U32;
typedef signed short S16;


// these constants determine the order in which 
// the cuboid is filled: 
#define DX 12
#define DY 3
#define DZ 1


// global verbose flag
bool verbose = false;


// --- Position class --------------------------------------------------------------------
// (represents a single piece in a certain translated and rotated state in the target cube)
class Position {
public:
    // 60 bits, one for each unit
    U64 v;

    // raw constructor
    Position(U64 v = 0): v(v) {}
    
    // piece constructor
    // (xyzPieceDesc: x0, y9, z0, x1, y1, z1, ..., -1)
    Position(int *&xyzPieceDesc):v(0) {
	for(; xyzPieceDesc[0] >= 0; xyzPieceDesc += 3)
	    set(xyzPieceDesc[0], xyzPieceDesc[1], xyzPieceDesc[2]);
	xyzPieceDesc++; // skip trailing -1
    }
    
    // plane constructor (use -1 for dont care coordinates)
    Position(int xx, int yy, int zz):v(0) {
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 4; y++) {
		for(int x = 0; x < 5; x++) {
		    if(((xx < 0) || (x == xx)) &&
		       ((yy < 0) || (y == yy)) &&
		       ((zz < 0) || (z == zz)))
			set(x, y, z);
		}
	    }
	}
    }
	
    // set single dot
    void set(int x, int y, int z) {
	assert(x < 5);
	assert(y < 4);
	assert(z < 3);
	v |= 1ull << (x * DX + y * DY + z * DZ);
    }
    
    // get single dot
    bool get(int x, int y, int z) const {
	assert(x < 5);
	assert(y < 4);
	assert(z < 3);
	return (v & (1ull << (x * DX + y * DY + z * DZ))) != 0;
    }
    
    // rotation around the Z axis (90 deg)
    void rotZ() {
	Position n;
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 4; y++) {
		for(int x = 0; x < 4; x++) {
		    if(get(x, y, z))
			n.set(3 - y, x, z);
		}
	    }
	}	
	v = n.v;
    }

    // rotation around the Y axis (90 deg)
    void rotY() {
	Position n;
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 4; y++) {
		for(int x = 0; x < 3; x++) {
		    if(get(x, y, z))
			n.set(2 - z, y, x);
		}
	    }
	}	
	v = n.v;
    }

    // rotation around the Y axis (180 deg) (left aligned)
    void rotY2() {
	Position n;
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 4; y++) {
		for(int x = 0; x < 4; x++) {
		    if(get(x, y, z))
			n.set(3 - x, y, 2 - z);
		}
	    }
	}	
	v = n.v;
    }
    
    // rotation around the X axis (90 deg)
    void rotX() {
	Position n;
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 3; y++) {
		for(int x = 0; x < 5; x++) {
		    if(get(x, y, z))
			n.set(x, 2 - z, y);
		}
	    }
	}	
	v = n.v;
    }
    
    // mirror on x, y or z axis
    void mirror(bool xx, bool yy, bool zz) {
	Position n;
	for(int z = 0; z < 3; z++) {
	    for(int y = 0; y < 4; y++) {
		for(int x = 0; x < 5; x++) {
		    if(get(x, y, z))
			n.set(xx ? 4 - x : x, yy ? 3 - y : y, zz ? 2 - z : z);
		}
	    }
	}	
	v = n.v;
    }
    
    // rotation around the Z axis (180 deg)
    void rotZ180() { mirror(1, 1, 0); }
    
    // rotation around the Y axis (180 deg)
    void rotY180() { mirror(1, 0, 1); }

    // rotation around the X axis (180 deg)
    void rotX180() { mirror(0, 1, 1); }

    // mirror Z axis
    void mirrorZ() { mirror(0, 0, 1); }

    // mirror Y axis
    void mirrorY() { mirror(0, 1, 0); }
    
    // mirror X axis
    void mirrorX() { mirror(1, 0, 0); }

    // get number of bits set
    int getNumBitsSet() {
	int r = 0;
	for(int i = 0;i < 60; i++) {
	    if(v & (1ull << i)) r++;
	}
	return r;
    }
    
    // print
    void print(int n, const char *piece) {
	printf("n=%3d piece=%s bits=%2d v=%015llX\n", n, piece, getNumBitsSet(), v);
	for(int y = 3; y >= 0; y--) {
	    printf("y=%d:", y);
	    for(int z = 0; z < 3; z++) {
		printf("  ");
		for(int x = 0; x < 5; x++) {
		    printf("%c", get(x, y, z) ? '*' : '-');
		}
	    }
	    printf("\n");
	}
    }
};


// --- these planes are used while generating the translations and rotations below ------------------------------
Position x0(0, -1, -1), x3(3, -1, -1), x4(4, -1, -1), Y0(-1, 0, -1), y3(-1, 3, -1), z0(-1, -1, 0), z2(-1, -1, 2);
// (unbelievable namespace pollution: y0 is declared as global symbol in some system header file!)


// --- PosList class ------------------------------------------------------------
// (represents all possible translations/rotations of one of the 12 pieces)
class PosList {
public:
    // list of positions
    vector<Position> pos;
    
    // default constructor
    PosList() {}

    // generating constructor
    // (generates all possible rotations and translations of a single piece)
    PosList(const Position &p) {
	rotate(p);
    }

    // check whether a specific position is already present
    bool hasPos(const Position &p) {
	for(size_t i = 0; i < pos.size(); i++)
	    if(pos[i].v == p.v) 
		return true;
	return false;
    }
    
    // add position, ignored if position already in list
    bool add(const Position &p) {
	if(hasPos(p)) 
	    return true;
	pos.push_back(p);
	return false;
    }
    
    // add all possible translations of position
    void translate(const Position &p) {
	if(add(p)) return;
	if(!(p.v & x0.v)) translate(p.v >> DX);
	if(!(p.v & x4.v)) translate(p.v << DX);
	if(!(p.v & Y0.v)) translate(p.v >> DY);
	if(!(p.v & y3.v)) translate(p.v << DY);
	if(!(p.v & z0.v)) translate(p.v >> DZ);
	if(!(p.v & z2.v)) translate(p.v << DZ);
    }

    // add all possible Y axis rotations and translations
    void rotateY(const Position &q) {
	Position p = q;
	translate(p);
	
	// left align small pieces to allow Y rotation
	if(!(p.v & x0.v)) p.v >>= DX;
	
	// large pieces can only have two orientations
	if(p.v & x3.v) {
	    p.rotY2();
	    translate(p);
	    p.rotY2();
	} else {
	    p.rotY();
	    translate(p);
	    p.rotY();
	    translate(p);
	    p.rotY();
	    translate(p);
	    p.rotY();
	}
    }
    
    // add all possible rotations and translations
    void rotate(const Position &q) {
	Position p = q;
	rotateY(p);
        p.rotZ();
	rotateY(p);
        p.rotZ();
	rotateY(p);
        p.rotZ();
	rotateY(p);
        p.rotZ();
	assert(p.v == q.v);
	p.rotX();
	rotateY(p);
        p.rotX();
        p.rotX();
	rotateY(p);
        p.rotX();
	assert(p.v == q.v);
    }
    
    // eliminate rotation symmetry for this piece (x, y and z)
    // (choose variants which have lower bits set)
    void eliminateSymmetry() {
	vector<Position> oldpos = pos;
	pos.clear();
	for(size_t i = 0; i < oldpos.size(); i++) {
	    Position px = oldpos[i];
	    Position py = oldpos[i];
	    Position pz = oldpos[i];
	    px.rotX180();
	    py.rotY180();
	    pz.rotZ180();
	    size_t j;
	    for(j = 0; j < pos.size(); j++) {
		if((pos[j].v == px.v) || (pos[j].v == py.v) || (pos[j].v == pz.v)) {
		    if(oldpos[i].v < pos[j].v)
			pos[j].v = oldpos[i].v;  // choose variant with lower bits
		    break;
		}
	    }
	    if(j == pos.size()) 
		pos.push_back(oldpos[i]);
	}
    }

    // eliminate symmetry for this piece (y only)
    // (choose variants which have higher bits set)
    void eliminateYSymmetry() {
	vector<Position> oldpos = pos;
	pos.clear();
	for(size_t i = 0; i < oldpos.size(); i++) {
	    Position py = oldpos[i];
	    py.rotY180();
	    size_t j;
	    for(j = 0; j < pos.size(); j++) {
		if(pos[j].v == py.v) {
		    // thanks to epsilon0 for this simple expression to get
		    // the lowest bit set (a & -a)
		    if((oldpos[i].v & -oldpos[i].v) > (pos[j].v & -pos[j].v))
			pos[j].v = oldpos[i].v; // choose variant with higher bits
		    break;
		}
	    }
	    if(j == pos.size()) 
		pos.push_back(oldpos[i]);
	}
    }

    // eliminate mirror symmetry for this piece (x, y and/or z)
    // (mSymEliminated[pos] is set to true/false if symmetry was eliminated/not eliminated)
    void eliminateMSymmetry(bool elimX, bool elimY, bool elimZ, bool *mSymEliminated = 0) {
	vector<Position> oldpos = pos;
	pos.clear();
	for(size_t i = 0; i < oldpos.size(); i++) {
	    Position px = oldpos[i];
	    Position py = oldpos[i];
	    Position pz = oldpos[i];
	    if(elimX)
		px.mirrorX();
	    if(elimY)
		py.mirrorY();
	    if(elimZ)
		pz.mirrorZ();
	    size_t j;
	    for(j = 0; j < pos.size(); j++) {
		if((pos[j].v == px.v) || (pos[j].v == py.v) || (pos[j].v == pz.v)) {
		    if((oldpos[i].v & -oldpos[i].v) > (pos[j].v & -pos[j].v)) {
			pos[j].v = oldpos[i].v;
		    }
		    if(mSymEliminated)
			mSymEliminated[j] = true;
		    break;
		}
	    }
	    if(j == pos.size()) {
		if(mSymEliminated)
		    mSymEliminated[j] = false;
		pos.push_back(oldpos[i]);
	    }
	}
    }

    // check all positions (just a consistency check)
    void checkNumBits(int num, const char *piece) {
	for(size_t i = 0; i < pos.size(); i++) {
	    if(pos[i].getNumBitsSet() != num) {
		// error
		printf("checkNumBits: pos[i].getNumBitsSet() != num: %d != %d\n", pos[i].getNumBitsSet(), num);
		pos[i].print(i, piece);
		exit(1);
	    }
	}
    }
    
    // print all positions for this piece
    void print(const char *piece) {
	for(size_t i = 0; i < pos.size(); i++) {
	    pos[i].print(i, piece);
	}
    }
};


// --- global data used during the main calculation --------------------------

// positions for 12 pieces
PosList poslist[12];

// number of solutions found so far
int numsol = 0;

// number of pieces laid so far (number of tries)
U32 tests = 0;

// stuff for time measurement
double starttime;
clock_t lastClock;
double curtime;

// number of bits set for each of the 12 pieces
U32 piece2numbits[12];

// 60 bit masks for all positions (4 to 6 bits set)
U64 posinfo_occ[3500];

// 12 bit masks for all positions (1 bit set each, shifted left by 6)
U32 posinfo_piecemask[3500];

// index table for piece lookup table
U32 listindex[1 << (6 + 12)];

// piece lookup tables
S16 list[8*1024*1024];

// bit masks (one bit set each)
U64 bits[64];

// function prototypes
void foundSolution(U64 occupied, U32 nextfree, U32 pieceslaid);
void done();



// --- precalculation --------------------------------------------------------
// (calculation of bit patterns and lookup tables)
void precalculation(int selectPiece, int selectPiecePos, bool elimYsym, bool elimX, bool mirror) {
    // this table contains the data for the 12 pieces
    int pieces[] = {
    'c', 5, 0,0,0, 1,0,0, 2,0,0, 0,1,0, 2,1,0, -1,
    'f', 5, 1,0,0, 0,1,0, 1,1,0, 2,1,0, 0,2,0, -1,
    't', 6, 0,0,0, 1,0,0, 2,0,0, 3,0,0, 1,1,0, 3,1,0, -1,
    'T', 5, 1,0,0, 1,1,0, 0,2,0, 1,2,0, 2,2,0, -1,
    's', 5, 2,0,0, 0,1,0, 1,1,0, 2,1,0, 0,2,0, -1,
    'p', 5, 0,0,0, 1,0,0, 2,0,0, 0,1,0, 1,1,0, -1,
    'q', 5, 0,0,0, 1,0,0, 0,1,0, 1,1,0, 0,1,1, -1,
    'x', 5, 1,0,0, 0,1,0, 1,1,0, 2,1,0, 1,2,0, -1,
    '4', 4, 0,0,0, 1,0,0, 0,1,0, 0,1,1, -1,
    'l', 5, 0,0,0, 1,0,0, 2,0,0, 3,0,0, 0,1,0, -1,
    'w', 5, 0,0,0, 0,1,0, 1,1,0, 1,2,0, 2,2,0, -1,
    'y', 5, 0,0,0, 1,0,0, 2,0,0, 3,0,0, 1,1,0, -1};
    
    // generate rotations/translations (positions) for 12 pieces
    if(verbose)
	printf("generating translations/rotations for pieces:\n");
    int *ip = pieces;
    int total = 0;
    int symkill = 6; // q
    for(int i = 0; i < 12; i++) {
	// create all positions for piece
	char name[2];
	name[0] = *(ip++);
	name[1] = 0;
	int numBits = *(ip++);
	piece2numbits[i] = numBits;
	poslist[i] = PosList(Position(ip));	
	if(verbose)
	    printf("number of translations/rotations for piece %2d (%s): %3u\n", i, name, unsigned(poslist[i].pos.size()));

	// one piece is used to kill the y rotation symmetry or the X mirror symmetry
	if(i == symkill) {
	    if(elimYsym) {
		poslist[i].eliminateYSymmetry();
		if(verbose)
		    printf("number of translations/rotations for piece %2d after eliminating Y rotation symmetry: %3u\n", i, unsigned(poslist[i].pos.size()));
	    }
	    if(elimX) {
		poslist[i].eliminateMSymmetry(elimX, false, false);
		if(verbose)
		    printf("number of translations/rotations for piece %2d after eliminating X mirror symmetry: %3u\n", i, unsigned(poslist[i].pos.size()));
	    }
	}
	
	// one piece is locked in the cuboid to reduce complexity
	if(i == selectPiece) {
	    poslist[i].eliminateSymmetry();
	    poslist[selectPiece].eliminateMSymmetry(true, true, true);
	    poslist[i].pos[0] = poslist[i].pos[selectPiecePos];
	    poslist[i].pos.resize(1);
	    if(verbose)
		printf("number of translations/rotations for piece %2d after selecting pos %d: %3u\n", i, selectPiecePos, unsigned(poslist[i].pos.size()));
	}

	// one piece is not mirror symmetric and can be used to introduce a mirror symmetry
	if(i == 8) {
	    if(mirror) {
		vector<Position> pos;
		for(size_t j = 0; j < poslist[i].pos.size(); j++) {
		    Position p = poslist[i].pos[j];
		    p.mirrorZ();
		    pos.push_back(p);
		}
		poslist[i].pos.insert(poslist[i].pos.end(), pos.begin(), pos.end());
		if(verbose)
		    printf("number of translations/rotations for piece %2d after mirroring: %3u\n", i, unsigned(poslist[i].pos.size()));
	    }
	}

	// generate posinfo lists
	for(size_t j = 0; j < poslist[i].pos.size(); j++, total++) {
	    posinfo_occ[total] = poslist[i].pos[j].v;
	    posinfo_piecemask[total] = 1 << (i + 6);
	}	    
	
	// check consistency
	poslist[i].checkNumBits(numBits, name);
    }
    if(verbose)
	printf("%d positions total\n", total);

    // stop here for initialization
    if(selectPiece < 0) return;
    
    // generate 'bits' array 
    memset(bits, 0, sizeof(bits));
    U64 forbid_occ = poslist[selectPiece].pos[0].v;
    int k = 0;
    for(int i = 0; i < 60; i++) {
	if((forbid_occ & (1ull << i)) == 0) {
	    bits[k] = 1ull << i;
	    k++;
	}
    }
    
    // these array are used to speed up the lookup table generation process
    U32 pieces2minbits[4096];
    U32 pieces2maxbits[4096];
    for(U32 i = 0; i < 4096; i++) {
	pieces2minbits[i] = 0;
	pieces2maxbits[i] = 0;
	for(U32 j = 0; j < 12; j++) {
	    if(i & (1 << j)) {
		pieces2minbits[i]++;
		pieces2maxbits[i] += piece2numbits[j];
	    }	    
	}
    }
    
    // generate lookup tables
    U32 index = 0;
    list[index++] = -1;
    U32 forbid_piecemask = 1 << (selectPiece+6);
    if(verbose) {
	printf("generating lookup tables:\n");
	printf("%s\r", "............................................................" + piece2numbits[selectPiece]);
	fflush(stdout);
    }
    U64 mask = forbid_occ;
    U32 numlists = 0;    
    U32 numsaved = 0;
    memset(listindex, 0, sizeof(listindex));
    // for all bit positions
    for(U32 bitpos = 0; bitpos < (60 - piece2numbits[selectPiece]); bitpos++) {
	U64 value = bits[bitpos];
	mask |= value;
	// for all combinations of pieces laid
	for(U32 pieces = 0; pieces < (0x1000 << 6); pieces += 0x40) {
	    // do not calculate what we already have
	    if(listindex[bitpos | pieces]) {
		numsaved++;
		continue;
	    }
	    // do not calculate if pieceslaid combination does not make sense for bit position
	    if((pieces & forbid_piecemask) == 0) continue;
	    if(pieces2minbits[(pieces & ~forbid_piecemask) >> 6] > bitpos) continue;
	    if(pieces2maxbits[(pieces & ~forbid_piecemask) >> 6] < bitpos) continue;
	    
	    // generate list of pieces which can be placed at bitpos given that
	    // all bits up to bitpos are filled and given that 'pieces' pieces are
	    // not available anymore
	    U32 oldindex = index;
	    for(int pos = 0; pos < total; pos++) {
		if((pieces & posinfo_piecemask[pos]) == 0) {
		    if((posinfo_occ[pos] & mask) == value) {
			list[index++] = pos;
		    }
		}
	    }
	    
	    if(index == oldindex) {
		// ignore empty lists		
		listindex[bitpos | pieces] = 0;
	    } else {
		// this logic reuses all possible end parts of the currently generated list
		// so we save some memory
		U32 addpieces = 0;
		for(U32 i = oldindex; i < index; i++) {
		    listindex[bitpos | pieces | addpieces] = i;
		    addpieces |= posinfo_piecemask[list[i]];
		    while((i < (index - 1)) && (posinfo_piecemask[list[i]] == posinfo_piecemask[list[i + 1]])) i++;
		}
		// terminate list with -1
		list[index++] = -1;
		numlists++;
	    }
	}
	if(verbose) {
	    printf("*");
	    fflush(stdout);
	}
    }
    if(verbose)
	printf("\n%d/%d (%d%%) of static list entries used (%d/%d lists total (%d%%)) (numsaved=%d)\n\n", index, sizeof(list) / sizeof(list[0]), (index * 100)/(sizeof(list) / sizeof(list[0])), numlists, 60*4096, numlists*100/(60*4096), numsaved);
}


// --- get current time in seconds -------------------------------------------------------
double gettime() {
    // set this to '1' of you have a very slow CPU or if you dont have 'clock()'
#if 0    
    struct timeval t;
    gettimeofday(&t, 0);
    return double(t.tv_sec) + double(t.tv_usec)/1000000.0;
#else
    clock_t clk = clock();
    curtime += double(clk - lastClock) / double(CLOCKS_PER_SEC);
    lastClock = clk;
    return curtime;
#endif
}


// --- main calculation --------------------------------------------------------------------
// (recursively try out all possible combinations which fill up the target cuboid bit by bit
// in order (without leaving any holes))
//   occupied: bits already occupied
//   nextfree: index of next zero bit
//   pieceslaid: pieces already laid 
void solve(U64 occupied, U32 nextfree, U32 pieceslaid) {
    // for all positions which can be placed onto next free bit
    for(int i = listindex[pieceslaid | nextfree]; list[i] >= 0; i++) {
	// can piece be laid down?
	if((occupied & posinfo_occ[list[i]]) == 0) {
//	    tests++;
	    U64 o = occupied | posinfo_occ[list[i]];
	    U32 p = pieceslaid | posinfo_piecemask[list[i]];
	    // find next free bit
	    // (the loop usually only counts very few bits)
	    U32 n = nextfree;
	    for(n++; o & bits[n]; n++);
	    // solution found?
	    // (5 bits already taken by 'X' piece)
	    if(n != 55) {
		// no: try to fill more bits recursively
		solve(o, n, p);
	    } else {
		// found solution
		foundSolution(o, n, p);
		// no need to try more positions for this bit
		return;
	    } 
	}
    }
}


// --- this function is called whenever a solution is found ------------------------------
void foundSolution(U64 occupied, U32 nextfree, U32 pieceslaid) {
    // count solution
    numsol++;
    
    // sanity check
    if((pieceslaid != (0xfff << 6)) || (occupied != 0xfffffffffffffffull) || (nextfree != 55)) {
	printf("\ninternal error: erroneous solution!\n");
	exit(1);
    }
    
    // display progress
    if((numsol & 0xfff) == 0)
    {
	double sec = gettime() - starttime;
	printf("%6d solutions found (tests=%.1fM, tests/sec=%.3fM, solutions/sec=%7.2f)  \r",
	       numsol, double(tests)/1000000.0, double(tests)/sec/1000000, double(numsol)/sec);
	fflush(stdout);
    }
}


// --- print summary ----------------------------------------------------------------------
void done() {
    double sec = gettime() - starttime;
    printf("%6d solutions found (tests=%d (%.1fM), tests/sec=%.3fM, solutions/sec=%7.2f, seconds=%6.2f (%.0f:%05.2f))\n",
	   numsol, tests, double(tests)/1000000.0, double(tests)/sec/1000000.0, double(numsol)/sec, sec, floor(sec/60.0), sec - floor(sec/60.0)*60.0);
}


// --- main -----------------------------------------------------------------
int main(int argc, char *argv[])
{
    // check parameters
    printf("ctpuzzle solver *** V1.1 (C) 2003 by Johannes Overmann <Johannes.Overmann@gmx.de>\n");
    for(int i = 1; i < argc; i++) {
	if((!strcmp(argv[i], "--verbose")) || (!strcmp(argv[i], "-v"))) {
	    verbose = true;
	} else {
	    printf("Usage: %s [--verbose|-v]\n", argv[0]);
	    exit(0);
	}
    }
    
    // initialize clock
    lastClock = clock();
    starttime = gettime();
    
    // the x piece (nr 7) is chosen to be laid first
    // this reduces the complexity of the problem
    int selectPiece = 7; 
    precalculation(-1, -1, false, false, false);
    // eliminate rotation symmetry for 'X' piece positions    
    poslist[selectPiece].eliminateSymmetry();
    // eliminate mirror symmetry for 'X' piece positions
    bool mSymEliminated[1000];
    poslist[selectPiece].eliminateMSymmetry(true, true, true, mSymEliminated);
    bool eliminateYSymmetry[1000];
    bool eliminateX[1000];
    bool mirror[1000];
    U64 occ[1000];
    // get number of remaining positions for 'X'
    int numPos = poslist[selectPiece].pos.size();
    // find out what exactly we must do for all remaining 'X' positions
    // to get rid of the rotation symmetry and to compensate the lost
    // mirror symmetry
    for(int i = 0; i < numPos; i++) {
	U64 v = poslist[selectPiece].pos[i].v;
	Position p = poslist[selectPiece].pos[i];
	if(verbose)
	    p.print(i, "x");
	// remember bit pattern
	occ[i] = p.v;
	// find out whether we need to eliminate rotation around the Y axis
	p.rotY180();
	eliminateYSymmetry[i] = p.v == v;
	// find out whether we need to eliminate mirror symmetry along the x axis
	p = poslist[selectPiece].pos[i];
	p.mirrorX();	
	eliminateX[i] = (!eliminateYSymmetry[i]) && (!mSymEliminated[i]) && (p.v == v);
	// find out whether we need to compensate for eliminated mirror symmetry
	mirror[i] = (!eliminateYSymmetry[i]) && (eliminateX[i] || mSymEliminated[i]);

	if(verbose) {
	    printf("%2d: %015llX, elimYsym=%d, mSymEliminated=%d, eliminateX=%d, mirror=%d\n",
		   i, v, eliminateYSymmetry[i], mSymEliminated[i], eliminateX[i], mirror[i]);
	}
    }
    
    // for all remaining positions for the 'X' piece call solve() and count solutions
    for(int i = 0; i < numPos; i++) {
	// recalculate new lookup tables for current 'X' position
	precalculation(selectPiece, i, eliminateYSymmetry[i], eliminateX[i], mirror[i]);
	// get next free bit
	U32 nextfree = 0;
	while(occ[i] & bits[nextfree]) nextfree++;
	// recursively solve puzzle
	if(verbose)
	    printf("solve(%015llX, %2d, %03X);\n", occ[i], nextfree, 1 << selectPiece);
	U32 _tests = tests;
	U32 _numsol = numsol;
	solve(occ[i], nextfree, 1 << (selectPiece + 6));
	_tests = tests - _tests;
	_numsol = numsol - _numsol;
	if(verbose) {
	    printf("solved %2d: numsol=%9d tests=%9d tests/numsol=%9.3f               \n", i, _numsol, _tests, double(_tests)/double(_numsol));
	    done();
	}
    }

    // print final statistics
    done();
    
    // success
    return 0;
}

